from typing import Iterator, Iterable, TypeAlias
from contextlib import contextmanager
import ast
from dataclasses import dataclass

from uncalled import whitelist

is_framework = whitelist.get_matcher()


# TODO: import as alias, not use


class Kind:
    MODULE = "module"
    CLASS = "class"
    FUNC = "function"
    NAME = "variable"


@dataclass(frozen=True)
class Namespace:
    kind: str
    name: str
    lineno: int


XPath: TypeAlias = tuple[Namespace, ...]


class Collector(ast.NodeVisitor):
    xpath: XPath
    definitions: list[XPath]
    references: list[XPath]

    def __init__(self) -> None:
        self.definitions = []
        self.references = []

    def visit_AsyncFunctionDef(self, fd: ast.AsyncFunctionDef):
        with self.enter_definition(Kind.FUNC, fd.name, fd.lineno):
            self.generic_visit(fd)

    def visit_FunctionDef(self, fd: ast.FunctionDef | ast.AsyncFunctionDef):
        with self.enter_definition(Kind.FUNC, fd.name, fd.lineno):
            self.generic_visit(fd)

    def visit_ClassDef(self, cd: ast.ClassDef):
        with self.enter_definition(Kind.CLASS, cd.name, cd.lineno):
            self.generic_visit(cd)

    @contextmanager
    def enter_definition(self, kind, name, lineno) -> Iterator[None]:
        if self.xpath[-1].kind is Kind.CLASS:
            name = "." + name
        self.xpath += (Namespace(kind, name, lineno),)
        self.definitions.append(self.xpath)
        yield
        self.xpath = self.xpath[:-1]

    def visit_Attribute(self, attr: ast.Attribute) -> None:
        name = attr.attr
        value = attr.value
        lineno = attr.lineno
        if isinstance(attr.ctx, ast.Store):
            if isinstance(value, ast.Name) and value.id == "self":
                namespace = Namespace(Kind.NAME, "." + name, lineno)
                self.definitions.append(self.xpath[:-1] + (namespace,))
                self.visit(value)
        else:
            self.references.append(self.xpath + (Namespace(Kind.NAME, name, lineno),))
            self.references.append(
                self.xpath + (Namespace(Kind.NAME, "." + name, lineno),)
            )
            self.visit(value)

    def visit_Name(self, name: ast.Name) -> None:
        id = name.id
        if isinstance(name.ctx, ast.Store):
            with self.enter_definition(Kind.NAME, id, name.lineno):
                pass
        else:
            if self.xpath[-1].kind is Kind.CLASS:
                id = "." + id
            self.references.append(
                self.xpath + (Namespace(Kind.NAME, id, name.lineno),)
            )

    # ImportFrom(identifier? module, alias* names, int? level)
    def visit_ImportFrom(self, imp: ast.ImportFrom) -> None:
        # TODO: handle aliases
        for alias in imp.names:
            if imp.module is None:
                self.references.append(
                    (
                        Namespace(Kind.MODULE, alias.name, imp.lineno),
                        Namespace(Kind.NAME, alias.name, imp.lineno),
                    )
                )
            else:
                self.references.append(
                    (
                        Namespace(Kind.MODULE, imp.module + ".py", imp.lineno),
                        Namespace(Kind.NAME, alias.name, imp.lineno),
                    )
                )


def collect(filenames: Iterable[str]) -> tuple[list, list]:
    c = Collector()
    for module, filename in parse_modules(filenames):
        c.xpath = (Namespace(Kind.MODULE, filename, 0),)
        c.visit(module)
    return c.references, c.definitions


def find_unused(
    all_references: list[XPath], all_definitions_paths: list[XPath]
) -> set[XPath]:
    references = set[str]()
    while True:
        new_references = {
            xpath[-1].name
            for xpath in all_references
            if is_reachable(xpath[:-1], references)
        }
        if new_references <= references:
            break
        references.update(new_references)
        all_references = [
            xpath for xpath in all_references if xpath[-1].name not in references
        ]
    return {
        xpath for xpath in all_definitions_paths if not is_reachable(xpath, references)
    }


def is_reachable(xpath: XPath, references: set[str]) -> bool:
    for item in xpath:
        if (
            item.kind is not Kind.CLASS
            and item.kind is not Kind.MODULE
            and item.name not in references
            and not is_framework(item.name)
        ):
            return False
    return True


def username_xpath(xpath: XPath) -> tuple[str, str]:
    x1, x2 = xpath[-2], xpath[-1]
    k1, k2 = x1.kind, x2.kind
    if k1 is Kind.CLASS:
        pair = x1.name + x2.name
        if k2 is Kind.CLASS:
            return "inner class", pair
        if k2 is Kind.FUNC:
            return "method", pair
        if k2 is Kind.NAME:
            return "attribute", pair
        else:
            assert False, str(k2)
    if k2 == Kind.NAME:
        return "variable", x2.name
    if k2 == Kind.FUNC:
        return "function", x2.name
    if k2 == Kind.CLASS:
        return "class", x2.name
    assert False, str(x2)


def parse_modules(filenames: Iterable[str]) -> Iterator[tuple[ast.Module, str]]:
    for filename in filenames:
        with open(filename, encoding="utf-8") as f:
            source = f.read()
        try:
            module = ast.parse(source, filename=filename)
        except SyntaxError:
            from sys import stderr

            print("Could not parse " + filename, file=stderr)
        else:
            yield module, filename


def print_unused(names: set[XPath]) -> None:
    for xpath in sorted(names):
        if xpath[-1].kind is Kind.CLASS and not whitelist.Flags.track_classes:
            continue
        if xpath[-1].kind is Kind.NAME and not whitelist.Flags.track_variables:
            continue
        kind, fullname = username_xpath(xpath)
        module, *_, file = xpath
        print(f"{module.name}:{file.lineno}: Unused {kind} '{fullname}'")


def run(files: Iterable[str]) -> None:
    all_references, all_definitions_paths = collect(files)
    print_unused(find_unused(all_references, all_definitions_paths))



import re
from pathlib import Path

from uncalled.single_pass import Finder


def prepare(txt: str) -> str:
    # this is a hack to remove comments. It may remove too much, but rarely
    txt = re.sub(r"#[^\r\n]*", "", txt)

    return re.sub(r'"""[^\r]*?"""' + r"|'''[^\r]*?'''", "...", txt)


class RegexFinder(Finder):
    def __init__(self, filename: Path, txt: str) -> None:
        super().__init__(filename, txt)
        self.txt = prepare(txt)

    def find_defs(self) -> set[str]:
        return set(
            [
                x.strip()[4:]
                for x in re.findall(r"^\s+def [^\d\W]\w*", self.txt, re.MULTILINE)
            ]
        )

    def find_uses(self) -> set[str]:
        return set(re.findall(r"(?<!def )[^\d\W]\w*", self.txt, re.IGNORECASE))

    def find_prefixes(self) -> set[str]:
        return set(
            re.findall(r"""(?<=['"])[a-z_]+(?=['"]\s*[+])""", self.txt, re.IGNORECASE)
        )



import argparse
from uncalled import single_pass
from uncalled import ast_finder
from uncalled import regex_finder


def main():
    parser = argparse.ArgumentParser(
        description="Find uncalled function in Python projects"
    )
    parser.add_argument(
        "--how",
        choices=["ast", "regex", "both"],
        default="regex",
        help='technique to use. use "both" to reduce false positives [default: regex]',
    )
    parser.add_argument("files", nargs="+", default=".", help="files to analyze")
    args = parser.parse_args()

    results_ast = set()
    if args.how in ["both", "ast"]:
        results_ast = single_pass.run(args.files, ast_finder.AstFinder)
    results_regex = set()
    if args.how in ["both", "regex"]:
        results_regex = single_pass.run(args.files, regex_finder.RegexFinder)
    if args.how == "both":
        result = results_ast & results_regex
    else:
        result = results_ast | results_regex
    single_pass.report(result)


if __name__ == "__main__":
    main()



import os
import typing
from pathlib import Path
from typing import Iterable, Iterator

from uncalled.whitelist import get_matcher

is_framework = get_matcher("")


class Finder(typing.Protocol):
    def __init__(self, filename: Path, txt: str) -> None: ...
    def find_defs(self) -> set[str]: ...

    def find_uses(self) -> set[str]: ...

    def find_prefixes(self) -> set[str]: ...


def is_venv(filename: Path) -> bool:
    basename = os.path.basename(filename)
    return basename in ["venv", ".venv", "virtualenv", ".virtualenv"]


def read_file(paths: Iterable[Path]) -> Iterator[tuple[Path, str]]:
    for path in paths:
        basename = path.name
        if basename.startswith("."):
            continue
        if path.is_file() and path.suffix == ".py":
            with open(path, encoding="utf-8") as f:
                yield (path, f.read())
        elif path.is_dir():
            if basename.startswith("__"):
                continue
            if is_venv(path):
                continue
            yield from read_file([path / f for f in os.listdir(path)])


def run(
    filenames: Iterable[str], make_finder: typing.Callable[[Path, str], Finder]
) -> set[tuple[Path, str]]:
    # normalize filenames
    file_paths = [Path(f) for f in filenames]
    file_text = dict(read_file(file_paths))
    files = list(file_text.keys())
    finders = {f: make_finder(f, txt) for f, txt in file_text.items()}
    file_defs = {f: finders[f].find_defs() for f, txt in file_text.items()}
    file_uses = {f: finders[f].find_uses() for f, txt in file_text.items()}
    file_pref = {f: finders[f].find_prefixes() for f, txt in file_text.items()}
    prefs = [p for f, prefs in file_pref.items() for p in prefs]
    uses = {call for calls in file_uses.values() for call in calls}
    file_unused_defs = {
        (file, name)
        for file in files
        for name in file_defs[file] - uses
        if not is_framework(name) and not any(name.startswith(p) for p in prefs)
    }
    return file_unused_defs


def report(file_unused_defs: set[tuple[Path, str]]) -> None:
    for file, name in sorted(file_unused_defs):
        print("{}: Unused function {}".format(file, name))



import ast
import sys
from pathlib import Path

from uncalled.single_pass import Finder


class AstFinder(Finder):
    def __init__(self, filename: Path, txt: str) -> None:
        super().__init__(filename, txt)
        self.collector = StoreLoadCollector()
        try:
            parsed = ast.parse(txt)
            self.collector.generic_visit(parsed)
        except SyntaxError:
            print("Note: cannot parse ast of ", filename, file=sys.stderr)
            pass

    def find_defs(self) -> set[str]:
        return set(self.collector.definitions)

    def find_uses(self) -> set[str]:
        return set(self.collector.references)

    def find_prefixes(self) -> set[str]:
        return set()


class StoreLoadCollector(ast.NodeVisitor):
    def __init__(self) -> None:
        self.definitions: list[str] = []
        self.references: list[str] = []

    def visit_AsyncFunctionDef(self, fd: ast.AsyncFunctionDef) -> None:
        self.definitions.append(fd.name)
        self.generic_visit(fd)

    def visit_FunctionDef(self, fd: ast.FunctionDef) -> None:
        self.definitions.append(fd.name)
        self.generic_visit(fd)

    def visit_Attribute(self, attr: ast.Attribute) -> None:
        if isinstance(attr.ctx, ast.Store):
            pass
        else:
            self.references.append(attr.attr)
        self.generic_visit(attr.value)

    def visit_Name(self, name: ast.Name) -> None:
        if isinstance(name.ctx, ast.Store):
            pass
        else:
            self.references.append(name.id)

    def visit_Str(self, s: ast.Str) -> None:
        self.references.append(s.s)



import re


class Flags:
    include_strings = True
    ignore_underscored_methods = True
    ignore_underscored = True
    track_classes = False
    track_variables = True


class Frameworks:
    ast = True
    pytest = True
    unittest = True
    flask = True


def get_matcher(method_prefix=r'\.'):
    def methods(*items):
        return [method_prefix + x for x in items]

    prefixes = methods('__.+')
    if Frameworks.ast:
        prefixes += methods('generic_visit', 'visit_.+')
    if Frameworks.unittest:
        prefixes += ['tearDown', 'setUp']
    if Frameworks.pytest:
        prefixes += ['test_.+', 'call', 'pytest_.*']
        prefixes += methods('test_.+', 'runtest', 'run_test', 'set_up', 'setup', 'teardown', 'cases')
    if Frameworks.flask:
        prefixes += ['before_request', 'after_request', 'put', 'get', 'post', 'delete', 'patch',
                     'head', 'options', 'trace', 'route', 'errorhandler']
    if Flags.ignore_underscored_methods:
        prefixes += methods('_.+')
    if Flags.ignore_underscored:
        prefixes.append('_.+')

    return re.compile('|'.join('({})'.format(p) for p in prefixes)).fullmatch
